@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * Copyright (C) 2003 Imendio HB
+ * Copyright (C) 2003-2004 Imendio HB
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -28,23 +28,43 @@
 #include "gossip-roster.h"
 #include "gossip-marshal.h"
 #include "gossip-contact-info.h"
+#include "gossip-edit-groups-dialog.h"
 #include "gossip-stock.h"
 #include "gossip-utils.h"
 #include "gossip-roster-view.h"
 #include "gossip-log.h"
 
-#define d(x) 
+#define d(x)
+
+/* Flashing delay for icons (milliseconds). */
+#define FLASH_TIMEOUT 500
+
+/* 
+ * Active users are those which have recently changed state
+ * (e.g. online, offline or from normal to a busy state). 
+ * 
+ */
+
+/* time user is shown as active */
+#define ACTIVE_USER_SHOW_TIME 7000     
+        
+/* time after connecting which we wait before active users are enabled */
+#define ACTIVE_USER_WAIT_TO_ENABLE_TIME 5000    
+
 
 struct _GossipRosterViewPriv {
-	GtkTreeModel   *model;
-	GossipRoster   *roster;
+	GtkTreeModel        *model;
+	GossipRoster        *roster;
 
-	GtkItemFactory *item_popup_factory;
-	GtkItemFactory *group_popup_factory;
+	GtkItemFactory      *item_popup_factory;
+	GtkItemFactory      *group_popup_factory;
 
-	GHashTable     *flash_table;
+	GHashTable          *flash_table;
 
-	gboolean        show_offline;
+	gboolean             show_offline;
+	gboolean             show_active; /* show active users */
+
+	GtkTreeRowReference *drag_row;
 };
 
 typedef union {
@@ -57,98 +77,143 @@ typedef struct {
 	guint    flash_timeout_id;
 } FlashData;
 
-static void  roster_view_class_init         (GossipRosterViewClass *klass);
-static void  roster_view_init               (GossipRosterView      *view);
-static void  roster_view_finalize           (GObject               *object); 
-static GtkTreeModel * 
-roster_view_create_store                    (GossipRosterView      *view);
-static void roster_view_setup_tree          (GossipRosterView      *view);
-static void  roster_view_item_added         (GossipRoster          *roster,
-					     GossipRosterItem      *item,
-					     GossipRosterView      *view);
-static void  roster_view_item_updated       (GossipRoster          *roster,
-					     GossipRosterItem      *item,
-					     GossipRosterView      *view);
-static void  
-roster_view_item_presence_updated           (GossipRoster          *roster,
-					     GossipRosterItem      *item,
-					     GossipRosterView      *view);
-static void  roster_view_item_removed       (GossipRoster          *roster,
-					     GossipRosterItem      *item,
-					     GossipRosterView      *view);
-static void  roster_view_group_added        (GossipRoster          *roster,
-					     GossipRosterGroup     *group,
-					     GossipRosterView      *view);
-static void  roster_view_group_removed      (GossipRoster          *roster,
-					     GossipRosterGroup     *group,
-					     GossipRosterView      *view);
-static gint roster_view_iter_compare_func   (GtkTreeModel          *model,
-					     GtkTreeIter           *iter_a,
-					     GtkTreeIter           *iter_b,
-					     gpointer               user_data);
-static void  
-roster_view_pixbuf_cell_data_func           (GtkTreeViewColumn     *tree_column,
-					     GtkCellRenderer       *cell,
-					     GtkTreeModel          *tree_model,
-					     GtkTreeIter           *iter,
-					     GossipRosterView      *view);
-static void  
-roster_view_name_cell_data_func             (GtkTreeViewColumn     *tree_column,
-					     GtkCellRenderer       *cell,
-					     GtkTreeModel          *tree_model,
-					     GtkTreeIter           *iter,
-					     GossipRosterView      *view);
-static gboolean roster_view_find_group      (GossipRosterView      *view,
-					     GtkTreeIter           *iter,
-					     const gchar           *name);
-static gboolean roster_view_find_item       (GossipRosterView      *view,
-					     GtkTreeIter           *iter,
-					     GossipRosterItem      *item,
-					     GossipRosterGroup     *group);
-static gboolean 
-roster_view_button_press_event_cb           (GossipRosterView      *view,
-					     GdkEventButton        *event,
-					     gpointer               data);
-static void roster_view_row_activated_cb    (GossipRosterView      *view,
-					     GtkTreePath           *path,
-					     GtkTreeViewColumn     *col,
-					     gpointer               data);
-static void     
-roster_view_item_menu_remove_cb             (gpointer               data,
-					     guint                  action,
-					     GtkWidget             *widget);
-static void
-roster_view_item_menu_info_cb               (gpointer               data,
-					     guint                  action,
-					     GtkWidget             *widget);
-static void 
-roster_view_item_menu_log_cb               (gpointer               data,
-					     guint                  action,
-					     GtkWidget             *widget);
-static void 
-roster_view_item_menu_rename_cb             (gpointer               data,
-					     guint                  action,
-					     GtkWidget             *widget);
-static void 
-roster_view_group_menu_rename_cb            (gpointer               data,
-					     guint                  action,
-					     GtkWidget             *widget);
-static gchar *
-roster_view_item_factory_translate_func     (const gchar           *path,
-					     gpointer               data);
-static void roster_view_flash_free_data     (FlashData             *data);
-static void roster_view_add_item            (GossipRosterView      *view,
-					     GossipRosterItem      *item,
-					     GossipRosterGroup     *group);
-static void roster_view_remove_item         (GossipRosterView      *view,
-					     GossipRosterItem      *item,
-					     GossipRosterGroup     *group);
-static gboolean roster_view_iter_equal_item (GtkTreeModel          *model,
-					     GtkTreeIter           *iter,
-					     GossipRosterItem      *item);
+/* Drag & Drop types */
+enum DndDragType {
+	DND_DRAG_TYPE_JID,
+	NUM_DRAG_TYPES
+};
+
+typedef struct {
+	GossipRosterItem  *item;
+	GossipRosterView  *view;
+	GossipRosterGroup *group;
+
+	gboolean           remove;
+} ShowActiveData;
+
+static GtkTargetEntry drag_types[] = {
+	{ "text/jid", 0, DND_DRAG_TYPE_JID },
+};
+
+static GdkAtom drag_atoms[NUM_DRAG_TYPES];
+
+static void            roster_view_class_init                  (GossipRosterViewClass *klass);
+static void            roster_view_init                        (GossipRosterView      *view);
+static void            roster_view_finalize                    (GObject               *object);
+static GtkTreeModel *  roster_view_create_store                (GossipRosterView      *view);
+static void            roster_view_setup_tree                  (GossipRosterView      *view);
+static void            roster_view_connected_cb                (GossipApp             *app,
+								GossipRosterView      *view);
+static gboolean        roster_view_show_active_users_cb        (GossipRosterView      *view);
+static void            roster_view_item_added                  (GossipRoster          *roster,
+								GossipRosterItem      *item,
+								GossipRosterView      *view);
+static void            roster_view_item_updated                (GossipRoster          *roster,
+								GossipRosterItem      *item,
+								GossipRosterView      *view);
+static void            roster_view_item_presence_updated       (GossipRoster          *roster,
+								GossipRosterItem      *item,
+								GossipRosterView      *view);
+static gboolean        roster_view_item_active_cb              (ShowActiveData        *active);
+static ShowActiveData *roster_view_item_active_new             (GossipRosterView      *view,
+								GossipRosterItem      *item,
+								GossipRosterGroup     *group,
+								gboolean               remove);
+static void            roster_view_item_active_free            (ShowActiveData        *data);
+static void            roster_view_item_removed                (GossipRoster          *roster,
+								GossipRosterItem      *item,
+								GossipRosterView      *view);
+static void            roster_view_group_added                 (GossipRoster          *roster,
+								GossipRosterGroup     *group,
+								GossipRosterView      *view);
+static void            roster_view_group_removed               (GossipRoster          *roster,
+								GossipRosterGroup     *group,
+								GossipRosterView      *view);
+static gint            roster_view_iter_compare_func           (GtkTreeModel          *model,
+								GtkTreeIter           *iter_a,
+								GtkTreeIter           *iter_b,
+								gpointer               user_data);
+static void            roster_view_pixbuf_cell_data_func       (GtkTreeViewColumn     *tree_column,
+								GtkCellRenderer       *cell,
+								GtkTreeModel          *tree_model,
+								GtkTreeIter           *iter,
+								GossipRosterView      *view);
+static void            roster_view_name_cell_data_func         (GtkTreeViewColumn     *tree_column,
+								GtkCellRenderer       *cell,
+								GtkTreeModel          *tree_model,
+								GtkTreeIter           *iter,
+								GossipRosterView      *view);
+static void            roster_view_set_cell_background         (GossipRosterView      *view,
+								GtkCellRenderer       *cell,
+								gboolean               use_default);
+static gboolean        roster_view_find_group                  (GossipRosterView      *view,
+								GtkTreeIter           *iter,
+								const gchar           *name);
+static gboolean        roster_view_find_item                   (GossipRosterView      *view,
+								GtkTreeIter           *iter,
+								GossipRosterItem      *item,
+								GossipRosterGroup     *group);
+static gboolean        roster_view_button_press_event_cb       (GossipRosterView      *view,
+								GdkEventButton        *event,
+								gpointer               data);
+static void            roster_view_row_activated_cb            (GossipRosterView      *view,
+								GtkTreePath           *path,
+								GtkTreeViewColumn     *col,
+								gpointer               data);
+static void            roster_view_item_menu_remove_cb         (gpointer               data,
+								guint                  action,
+								GtkWidget             *widget);
+static void            roster_view_item_menu_info_cb           (gpointer               data,
+								guint                  action,
+								GtkWidget             *widget);
+static void            roster_view_item_menu_log_cb            (gpointer               data,
+								guint                  action,
+								GtkWidget             *widget);
+static void            roster_view_item_menu_rename_cb         (gpointer               data,
+								guint                  action,
+								GtkWidget             *widget);
+static void            roster_view_item_menu_edit_groups_cb    (gpointer               data,
+								guint                  action,
+								GtkWidget             *widget);
+static void            roster_view_group_menu_rename_cb        (gpointer               data,
+								guint                  action,
+								GtkWidget             *widget);
+static gchar *         roster_view_item_factory_translate_func (const gchar           *path,
+								gpointer               data);
+static void            roster_view_flash_free_data             (FlashData             *data);
+static void            roster_view_add_item                    (GossipRosterView      *view,
+								GossipRosterItem      *item,
+								GossipRosterGroup     *group);
+static gboolean        roster_view_remove_item_with_iter       (GossipRosterView      *view,
+								GtkTreeIter           *iter,
+								GossipRosterItem      *item,
+								GossipRosterGroup     *group);
+static void            roster_view_remove_item                 (GossipRosterView      *view,
+								GossipRosterItem      *item,
+								GossipRosterGroup     *group);
+static gboolean        roster_view_iter_equal_item             (GtkTreeModel          *model,
+								GtkTreeIter           *iter,
+								GossipRosterItem      *item);
+static void            roster_view_drag_begin                  (GtkWidget             *widget,
+								GdkDragContext        *context,
+								gpointer               user_data);
+static void            roster_view_drag_data_get               (GtkWidget             *widget,
+								GdkDragContext        *contact,
+								GtkSelectionData      *selection,
+								guint                  info,
+								guint                  time,
+								gpointer               user_data);
+static void            roster_view_drag_end                    (GtkWidget             *widget,
+								GdkDragContext        *context,
+								gpointer               user_data);
+
+
+
+
+
 
 enum {
-	ITEM_ACTIVATED,
+	CONTACT_ACTIVATED,
 	LAST_SIGNAL
 };
 
@@ -157,6 +222,7 @@ enum {
 	ITEM_MENU_REMOVE,
 	ITEM_MENU_INFO,
 	ITEM_MENU_RENAME,
+	ITEM_MENU_EDIT_GROUPS,
 	ITEM_MENU_LOG
 };
 
@@ -190,6 +256,14 @@ static GtkItemFactoryEntry item_menu_items[] = {
 		"<Item>",
 		NULL
 	},
+	{
+		N_("/_Edit groups"),
+		NULL,
+		GIF_CB (roster_view_item_menu_edit_groups_cb),
+		ITEM_MENU_EDIT_GROUPS,
+		"<Item>",
+		NULL
+	},
 	{
 		N_("/Show _Log"),
 		NULL,
@@ -239,7 +313,6 @@ static GtkTargetEntry row_targets[] = {
 };
 */
 
-GObjectClass *parent_class;
 static guint signals[LAST_SIGNAL];
 
 GType                
@@ -275,8 +348,8 @@ roster_view_class_init (GossipRosterViewClass *klass)
 
 	object_class->finalize = roster_view_finalize;
 
-	signals[ITEM_ACTIVATED] = 
-		g_signal_new ("item_activated",
+	signals[CONTACT_ACTIVATED] = 
+		g_signal_new ("contact_activated",
 			      G_TYPE_FROM_CLASS (klass),
 			      G_SIGNAL_RUN_LAST,
 			      0,
@@ -353,6 +426,30 @@ roster_view_init (GossipRosterView *view)
 			  "row_activated",
 			  G_CALLBACK (roster_view_row_activated_cb),
 			  NULL);
+
+	g_signal_connect (gossip_app_get (), "connected",
+			  G_CALLBACK (roster_view_connected_cb),
+			  view);
+}
+
+static void
+roster_view_connected_cb (GossipApp        *app, 
+			  GossipRosterView *view)
+{
+	/* set timeout to enable active users */
+	g_timeout_add (ACTIVE_USER_WAIT_TO_ENABLE_TIME, 
+		       (GSourceFunc)roster_view_show_active_users_cb,
+		       view);
+}
+
+static gboolean
+roster_view_show_active_users_cb (GossipRosterView *view)
+{
+	GossipRosterViewPriv *priv = view->priv;
+
+	priv->show_active = TRUE;
+
+	return FALSE;
 }
 
 static void
@@ -396,6 +493,7 @@ roster_view_setup_tree (GossipRosterView *view)
 	GossipRosterViewPriv *priv;
 	GtkCellRenderer      *cell;
 	GtkTreeViewColumn    *col;
+	static int            setup = 0;
 
 	priv = view->priv;
 	
@@ -421,7 +519,32 @@ roster_view_setup_tree (GossipRosterView *view)
 						 view,
 						 NULL);
 
+	if (!setup) {
+		int i;
+		
+		for (i = 0; i < NUM_DRAG_TYPES; ++i) {
+			drag_atoms[i] = gdk_atom_intern (drag_types[i].target,
+							 FALSE);
+		}
+
+		setup = 1;
+	}
+
 	gtk_tree_view_append_column (GTK_TREE_VIEW (view), col);
+
+	gtk_drag_source_set (GTK_WIDGET (view), GDK_BUTTON1_MASK, 
+			     drag_types, NUM_DRAG_TYPES, 
+			     GDK_ACTION_COPY);
+
+	g_signal_connect (GTK_TREE_VIEW (view), "drag-begin",
+			  G_CALLBACK (roster_view_drag_begin),
+			  NULL);
+	g_signal_connect (GTK_TREE_VIEW (view), "drag-data-get",
+			  G_CALLBACK (roster_view_drag_data_get),
+			  NULL);
+	g_signal_connect (GTK_TREE_VIEW (view), "drag-end",
+			  G_CALLBACK (roster_view_drag_end),
+			  NULL);
 }
 
 static void
@@ -443,7 +566,7 @@ roster_view_item_added (GossipRoster     *roster,
 	}
 
 	for (l = gossip_roster_item_get_groups (item); l; l = l->next) {
-		GossipRosterGroup *group = (GossipRosterGroup *) l->data;
+		GossipRosterGroup *group = l->data;
 
 		roster_view_add_item (view, item, group);
 	}
@@ -483,6 +606,7 @@ roster_view_item_updated (GossipRoster     *roster,
 
 		do {
 			GtkTreePath *path;
+			gboolean     last_child;
 
 			if (!roster_view_iter_equal_item (priv->model, &iter, item)) {
 				continue;
@@ -492,23 +616,33 @@ roster_view_item_updated (GossipRoster     *roster,
 			gtk_tree_model_row_changed (priv->model, path, &iter);
 			gtk_tree_path_free (path);
 
+			last_child = FALSE;
+			
 			for (l = gossip_roster_item_get_groups (item); l; l = l->next) {
-				GossipRosterGroup *group = (GossipRosterGroup *) l->data;
+				GossipRosterGroup *group = l->data;
 
 				if (strcmp (gossip_roster_group_get_name (e->group),
 					    gossip_roster_group_get_name (group)) == 0) {
 					continue;
 				}
-			} 
+				
+				if (!roster_view_remove_item_with_iter (view, &iter, item, e->group)) {
+					/* Ugly, but we need to break out of
+					 * both loops.
+					 */
+					last_child = TRUE;
+					break;
+				}
+			}
 
-			if (!gtk_tree_store_remove (GTK_TREE_STORE (priv->model), &iter)) {
+			if (last_child) {
 				break;
 			}
 		} while (gtk_tree_model_iter_next (priv->model, &iter));
 	} while (gtk_tree_model_iter_next (priv->model, &group_iter));
 	
 	for (l = gossip_roster_item_get_groups (item); l; l = l->next) {
-		GossipRosterGroup *group = (GossipRosterGroup *) l->data;
+		GossipRosterGroup *group = l->data;
 		GtkTreeIter        iter;
 
 		if (roster_view_find_item (view, &iter, item, group)) {
@@ -537,28 +671,78 @@ roster_view_item_presence_updated (GossipRoster     *roster,
 	if (priv->show_offline || !gossip_roster_item_is_offline (item)) {
 		show_item = TRUE;
 	}
-	
-	for (l = gossip_roster_item_get_groups (item); l; l = l->next) {
-		GossipRosterGroup *group = (GossipRosterGroup *) l->data;
-		
-		if (roster_view_find_item (view, &iter, item, group)) {
-			if (!show_item) {
-				/* Remove item */
-				d(g_print ("Remove item!\n"));
-				roster_view_remove_item (view, item, group);
-			} else { 
-				GtkTreePath *path;
-				
-				d(g_print ("Update item!\n"));
 
-				path = gtk_tree_model_get_path (priv->model,
-								&iter);
-				gtk_tree_model_row_changed (priv->model,
-							    path, &iter); 
+	for (l = gossip_roster_item_get_groups (item); l; l = l->next) {
+		GossipRosterGroup *group = l->data;
+
+		if (roster_view_find_item (view, &iter, item, group)) {
+			GtkTreePath *path;
+
+			if (!show_item) {
+				if (priv->show_active) {
+					ShowActiveData *data;
+
+					data = roster_view_item_active_new (view, 
+									    item, 
+									    group, 
+									    TRUE);
+
+					gossip_roster_item_set_active (item, TRUE);  
+					
+					g_timeout_add (ACTIVE_USER_SHOW_TIME, 
+						       (GSourceFunc) roster_view_item_active_cb,
+						       data);
+
+					/* update roster */
+					d(g_print ("Update item (show who has disconnected)!\n")); 
+					path = gtk_tree_model_get_path (priv->model, &iter);
+					gtk_tree_model_row_changed (priv->model, path, &iter); 
+					gtk_tree_path_free (path);						
+				} else {
+					/* Remove item */
+					d(g_print ("Remove item!\n"));
+					roster_view_remove_item (view, item, group);
+				}
+			} else { 
+				if (priv->show_active) {
+					ShowActiveData *data;
+					
+					data = roster_view_item_active_new (view, 
+									    item, 
+									    group, 
+									    FALSE);
+									
+					gossip_roster_item_set_active (item, TRUE);  
+					
+					g_timeout_add (ACTIVE_USER_SHOW_TIME, 
+						       (GSourceFunc) roster_view_item_active_cb,
+						       data);
+				}
+
+				/* update roster */
+				d(g_print ("Update item!\n"));
+				path = gtk_tree_model_get_path (priv->model, &iter);
+				gtk_tree_model_row_changed (priv->model, path, &iter); 
 				gtk_tree_path_free (path);
 			}
+
 			continue;
 		} else {
+			if (priv->show_active) {
+				ShowActiveData *data;
+
+				data = roster_view_item_active_new (view, 
+								    item, 
+								    group, 
+								    FALSE);
+				
+				gossip_roster_item_set_active (item, TRUE);   
+
+				g_timeout_add (ACTIVE_USER_SHOW_TIME,  
+					       (GSourceFunc) roster_view_item_active_cb, 
+					       data);
+			}
+
 			if (show_item) {
 				/* Add item */
 				d(g_print ("Add item!\n"));
@@ -570,6 +754,82 @@ roster_view_item_presence_updated (GossipRoster     *roster,
 	}
 }
 
+static ShowActiveData *
+roster_view_item_active_new (GossipRosterView  *view, 
+			     GossipRosterItem  *item,
+			     GossipRosterGroup *group,
+			     gboolean           remove)
+{
+	ShowActiveData *data;
+
+	g_return_val_if_fail (view != NULL, NULL);
+	g_return_val_if_fail (item != NULL, NULL);
+	g_return_val_if_fail (group != NULL, NULL);
+
+	data = g_new0 (ShowActiveData, 1);
+	
+	data->view = g_object_ref (view);
+	
+	data->item = gossip_roster_item_ref (item);
+	data->group = gossip_roster_group_ref (group);
+
+	data->remove = remove;
+
+	return data;
+}
+
+static void
+roster_view_item_active_free (ShowActiveData *data)
+{
+	g_return_if_fail (data != NULL);
+
+	g_object_unref (data->view);
+
+	gossip_roster_item_unref (data->item);
+	gossip_roster_group_unref (data->group);
+
+	g_free (data);
+}
+
+static gboolean
+roster_view_item_active_cb (ShowActiveData *data)
+{
+	GossipRosterViewPriv *priv;
+
+	g_return_val_if_fail (data != NULL, FALSE);
+
+	priv = data->view->priv;
+
+	if (data->remove &&
+	    gossip_roster_item_is_offline (data->item)) {
+		/* Remove item */
+		d(g_print ("Remove item!\n"));
+		roster_view_remove_item (data->view, 
+					 data->item, 
+					 data->group);
+	} else {
+		GtkTreeIter iter;
+
+		if (roster_view_find_item (data->view, 
+					   &iter, 
+					   data->item, 
+					   data->group)) {
+			GtkTreePath *path;
+
+			path = gtk_tree_model_get_path (priv->model, &iter);
+			gtk_tree_model_row_changed (priv->model, path, &iter); 
+			gtk_tree_path_free (path);
+		}
+	}
+
+	/* turn background highlight off */
+	gossip_roster_item_set_active (data->item, FALSE);  
+
+	roster_view_item_active_free (data);
+
+	return FALSE;
+}
+
 static void
 roster_view_item_removed (GossipRoster     *roster,
 			  GossipRosterItem *item,
@@ -583,37 +843,16 @@ roster_view_item_removed (GossipRoster     *roster,
 	d(g_print ("Item removed: %s\n", gossip_roster_item_get_name (item)));
 	
 	for (l = gossip_roster_item_get_groups (item); l; l = l->next) {
-		GossipRosterGroup *group = (GossipRosterGroup *) l->data;
-		GtkTreeIter        iter;
-		gboolean           is_group;
-		RosterElement     *e;
-		
-		if (!roster_view_find_item (view, &iter, item, group)) {
-			continue;
-		}
+		GossipRosterGroup *group = l->data;
 
-		gtk_tree_model_get (priv->model, &iter,
-				    COL_IS_GROUP, &is_group,
-				    COL_ELEMENT, &e,
-				    -1);
-
-		if (is_group) {
-			continue;
-		}
-
-		g_hash_table_remove (priv->flash_table, item);
-
-		gtk_tree_store_remove (GTK_TREE_STORE (priv->model), &iter);
-
-		gossip_roster_item_unref (e->item);
-		g_free (e);
+		roster_view_remove_item (view, item, group);
 	}
 }
 
 static void
-roster_view_group_added   (GossipRoster      *roster,
-			   GossipRosterGroup *group,
-			   GossipRosterView  *view)
+roster_view_group_added (GossipRoster      *roster,
+			 GossipRosterGroup *group,
+			 GossipRosterView  *view)
 {
 	GossipRosterViewPriv *priv;
 	GtkTreeIter           iter;
@@ -642,7 +881,6 @@ roster_view_group_added   (GossipRoster      *roster,
 	path = gtk_tree_model_get_path (priv->model, &iter);
 	gtk_tree_view_expand_row (GTK_TREE_VIEW (view), path, FALSE);
 	gtk_tree_path_free (path);
-
 }
 
 static void
@@ -746,19 +984,22 @@ roster_view_pixbuf_cell_data_func (GtkTreeViewColumn *tree_column,
 			    -1);
 
 	if (is_group) {
-		g_object_set (cell, "visible", FALSE, NULL);
+		g_object_set (cell, 
+			      "visible", FALSE,
+			      NULL);
+
+		roster_view_set_cell_background (view, cell, TRUE);
 		return;
 	} 
 
 	if (!gtk_tree_model_iter_parent (model, &parent, iter)) {
 		pixbuf = gossip_utils_get_pixbuf_from_stock (GOSSIP_STOCK_MESSAGE);
-	} else {	
+	} else {
 		flash = g_hash_table_lookup (priv->flash_table, e->item);
 		
 		if (flash && flash->flash_on) {
-			pixbuf = gossip_utils_get_pixbuf_from_stock (GOSSIP_STOCK_MESSAGE); 
-		}
-		else if (gossip_roster_item_is_offline (e->item)) {
+			pixbuf = gossip_utils_get_pixbuf_from_stock (GOSSIP_STOCK_MESSAGE);  
+		} else if (gossip_roster_item_is_offline (e->item)) {
 			pixbuf = gossip_utils_get_pixbuf_offline ();
 		} else {
 			GossipShow show = gossip_roster_item_get_show (e->item);
@@ -767,11 +1008,20 @@ roster_view_pixbuf_cell_data_func (GtkTreeViewColumn *tree_column,
 		}
 	}
 
+	if (gossip_roster_item_get_active (e->item)) {
+		roster_view_set_cell_background (view, cell, FALSE);
+	} else {
+		roster_view_set_cell_background (view, cell, TRUE);
+	}
+
 	g_object_set (cell,
 		      "visible", TRUE,
 		      "pixbuf", pixbuf,
 		      NULL);
-	g_object_unref (pixbuf);
+
+	if (pixbuf) {
+		g_object_unref (pixbuf);
+	}
 }
 
 static void
@@ -789,7 +1039,8 @@ ellipsize_string (gchar *str, gint len)
 	}
 }
 
-#define ELLIPSIS_LIMIT 6
+#define ELLIPSIS_MIN 6
+#define ELLIPSIS_MAX 100
 #define TREE_INDENT 30
 
 static void
@@ -807,10 +1058,12 @@ roster_view_ellipsize_item_string (GossipRosterView *view,
 	
 	len_str = g_utf8_strlen (str, -1);
 
-	if (len_str < ELLIPSIS_LIMIT) {
+	if (len_str < ELLIPSIS_MIN) {
 		return;
 	}
 
+	len_str = MIN (len_str, ELLIPSIS_MAX);
+
 	layout = gtk_widget_create_pango_layout (GTK_WIDGET (view), NULL);
 	
 	pango_layout_set_text (layout, str, -1);
@@ -833,7 +1086,7 @@ roster_view_ellipsize_item_string (GossipRosterView *view,
 		pango_attr_list_insert (attr_list, attr_size);
 	}
 	
-	while (len_str >= ELLIPSIS_LIMIT && width_str > width) {
+	while (len_str >= ELLIPSIS_MIN && width_str > width) {
 		len_str--;
 		ellipsize_string (str, len_str);
 		
@@ -853,6 +1106,45 @@ roster_view_ellipsize_item_string (GossipRosterView *view,
 	g_object_unref (layout);
 }
 
+
+static void
+roster_view_set_cell_background (GossipRosterView *view,
+			    GtkCellRenderer  *cell, 
+			    gboolean          use_default)
+{
+	GdkColor  color;
+	GtkStyle *style;
+	
+	g_return_if_fail (view != NULL);
+	g_return_if_fail (cell != NULL);
+
+	if (use_default) {
+		g_object_set (cell, 
+			      "cell-background-gdk", NULL, 
+			      NULL);
+	
+		return;
+	}
+
+	style = gtk_widget_get_style (GTK_WIDGET (view));
+
+/* 	color = style->base[GTK_STATE_SELECTED];  */
+/* 	color = style->text_aa[GTK_STATE_NORMAL];  */
+	color = style->bg[GTK_STATE_SELECTED];
+
+	/* Here we take the current theme colour and add it to
+	   the colour for white and average the two. This
+	   gives a colour which is inline with the theme but
+	   slightly whiter. */ 
+	color.red = (color.red + (style->white).red) / 2;
+	color.green = (color.green + (style->white).green) / 2;
+	color.blue = (color.blue + (style->white).blue) / 2;
+
+	g_object_set (cell, 
+		      "cell-background-gdk", &color, 
+		      NULL);
+}
+
 /* NOTE: We should write our own cell renderer instead of putting all these
  * nasty hacks here.
  */
@@ -890,9 +1182,17 @@ roster_view_name_cell_data_func (GtkTreeViewColumn *tree_column,
 			      "weight", PANGO_WEIGHT_BOLD,
 			      "text", gossip_roster_group_get_name (e->group),
 			      NULL);
+
+		roster_view_set_cell_background (view, cell, TRUE);
 		return;
 	} 
 
+	if (gossip_roster_item_get_active (e->item)) {	
+		roster_view_set_cell_background (view, cell, FALSE);
+	} else {
+		roster_view_set_cell_background (view, cell, TRUE);
+	}
+
 	/* FIXME: Figure out how to calculate the offset instead of
 	 * hardcoding it here (icon width + padding + indentation).
 	 */
@@ -954,6 +1254,7 @@ roster_view_name_cell_data_func (GtkTreeViewColumn *tree_column,
 	pango_attr_list_insert (attr_list, attr_style);
 
 	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (view));
+
 	if (!gtk_tree_selection_iter_is_selected (selection, iter)) {
 		attr_color = pango_attr_foreground_new (color.red, color.green, color.blue);
 		attr_color->start_index = attr_style->start_index;
@@ -1068,6 +1369,106 @@ roster_view_iter_equal_item (GtkTreeModel     *model,
 	return FALSE;
 }
 
+static void
+roster_view_drag_begin (GtkWidget      *widget,
+			GdkDragContext *context,
+			gpointer        user_data)
+{
+	GossipRosterViewPriv *priv;
+	GtkTreeSelection     *selection;
+	GtkTreeModel         *model;
+	GtkTreePath          *path;
+	GtkTreeIter           iter;
+
+	priv = GOSSIP_ROSTER_VIEW (widget)->priv;
+	
+	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (widget));
+	if (!gtk_tree_selection_get_selected (selection, &model, &iter)) {
+		return;
+	}
+
+	path = gtk_tree_model_get_path (model, &iter);
+	priv->drag_row = gtk_tree_row_reference_new (model, path);
+	gtk_tree_path_free (path);
+
+	/* FIXME: Set a nice icon */
+}
+
+static void
+roster_view_drag_data_get (GtkWidget             *widget,
+			   GdkDragContext        *contact,
+			   GtkSelectionData      *selection,
+			   guint                  info,
+			   guint                  time,
+			   gpointer               user_data)
+{
+	GossipRosterViewPriv *priv;
+	GtkTreePath          *src_path;
+	GtkTreeIter           iter;
+	GtkTreeModel         *model;
+	RosterElement        *e;
+	gboolean              is_group;
+	GossipJID            *jid;
+	const gchar          *str_jid;
+	
+	priv = GOSSIP_ROSTER_VIEW (widget)->priv;
+	
+	model = gtk_tree_view_get_model (GTK_TREE_VIEW (widget));
+	if (!priv->drag_row) {
+		return;
+	}
+
+	src_path = gtk_tree_row_reference_get_path (priv->drag_row);
+	if (!src_path) {
+		return;
+	}
+
+	if (!gtk_tree_model_get_iter (model, &iter, src_path)) {
+		gtk_tree_path_free (src_path);
+		return;
+	}
+
+	gtk_tree_path_free (src_path);
+
+	gtk_tree_model_get (model, &iter,
+			    COL_IS_GROUP, &is_group,
+			    COL_ELEMENT, &e,
+			    -1);
+	
+	if (is_group) {
+		return;
+	}
+
+	jid = gossip_roster_item_get_jid (e->item);
+	str_jid = gossip_jid_get_full (jid);
+	
+	switch (info) {
+	case DND_DRAG_TYPE_JID:
+		gtk_selection_data_set (selection, drag_atoms[info], 8, 
+					str_jid, strlen (str_jid) + 1);
+		break;
+	default:
+		return;
+	}
+	
+}
+
+static void
+roster_view_drag_end (GtkWidget      *widget,
+		      GdkDragContext *context,
+		      gpointer        user_data)
+{
+	GossipRosterViewPriv *priv;
+
+	priv = GOSSIP_ROSTER_VIEW (widget)->priv;
+
+	if (priv->drag_row) {
+		gtk_tree_row_reference_free (priv->drag_row);
+		priv->drag_row = NULL;
+	}
+}
+
+
 static gboolean
 roster_view_find_item (GossipRosterView  *view,
 		       GtkTreeIter       *iter,
@@ -1200,7 +1601,11 @@ roster_view_row_activated_cb (GossipRosterView  *view,
 	d(g_print ("Row activated!\n"));
 	
 	if (!is_group) {
-		g_signal_emit (view, signals[ITEM_ACTIVATED], 0, e->item);
+		GossipContact *contact;
+
+		contact = gossip_roster_get_contact_from_item (priv->roster,
+							       e->item);
+		g_signal_emit (view, signals[CONTACT_ACTIVATED], 0, contact);
 	}
 }
 
@@ -1372,6 +1777,23 @@ roster_view_item_menu_rename_cb (gpointer   data,
 	gossip_roster_rename_item (priv->roster, item, str);
 }
 
+static void
+roster_view_item_menu_edit_groups_cb (gpointer   data,
+				      guint      action,
+				      GtkWidget *widget)
+{
+	GossipRosterItem *item;
+
+	item = gossip_roster_view_get_selected_item (GOSSIP_ROSTER_VIEW (data));
+	if (!item) {
+		return;
+	}
+
+	gossip_edit_groups_new (gossip_roster_item_get_jid (item),
+				gossip_roster_item_get_name (item),
+				gossip_roster_item_get_groups (item));
+}
+
 static void 
 roster_view_group_menu_rename_cb (gpointer   data,
 				  guint      action,
@@ -1381,7 +1803,7 @@ roster_view_group_menu_rename_cb (gpointer   data,
 	GossipRosterViewPriv *priv;
 	gboolean              is_group;
 	RosterElement        *e;
-	gchar                *str;
+	gchar                *str, *tmp;
 	GtkTreeSelection     *selection;
 	GtkTreeIter           iter;
 	GtkTreeModel         *model;
@@ -1393,6 +1815,7 @@ roster_view_group_menu_rename_cb (gpointer   data,
 	priv = view->priv;
 	
 	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (view));
+
 	if (!gtk_tree_selection_get_selected (selection, &model, &iter)) {
 		return;
 	}
@@ -1406,8 +1829,9 @@ roster_view_group_menu_rename_cb (gpointer   data,
 		return;
 	}
 
-	str = g_strdup_printf ("<b>%s</b>", 
-			       gossip_roster_group_get_name (e->group));
+	tmp = g_markup_escape_text (gossip_roster_group_get_name (e->group), -1);
+	str = g_strdup_printf ("<b>%s</b>", tmp);
+	g_free (tmp);
 
 	/* Translator: %s denotes the group name */
 	dialog = gtk_message_dialog_new (GTK_WINDOW (gossip_app_get_window ()),
@@ -1476,6 +1900,8 @@ roster_view_flash_free_data (FlashData *data)
 	g_free (data);
 }
 
+#define EXPAND 1
+
 static void
 roster_view_add_item (GossipRosterView  *view,
 		      GossipRosterItem  *item,
@@ -1484,7 +1910,7 @@ roster_view_add_item (GossipRosterView  *view,
 	GossipRosterViewPriv *priv;
 	GtkTreeIter           iter, group_iter;
 	RosterElement        *e;
-#if 0
+#if EXPAND
 	gboolean              expand = FALSE;
 #endif
 
@@ -1493,41 +1919,52 @@ roster_view_add_item (GossipRosterView  *view,
 	d(g_print ("Adding item: [%s] to group: [%s]\n", 
 		   gossip_roster_item_get_name (item),
 		   gossip_roster_group_get_name (group)));
+
+	// koko
+	if (!priv->show_offline && gossip_roster_item_is_offline (item)) {
+		d(g_print ("Offline item: %s\n",
+			   gossip_roster_item_get_name (item)));
+		return;
+	}
 	
 	if (!group) {
 		gtk_tree_store_append (GTK_TREE_STORE (priv->model),
 				       &iter, 
 				       NULL);
-	}
-	else {
+	} else {
 		const gchar *name;
+
 		name = gossip_roster_group_get_name (group);
 
-		if (roster_view_find_group (view, &group_iter, name)) {
-			gtk_tree_store_append (GTK_TREE_STORE (priv->model),
-					       &iter,
-					       &group_iter);	
-		} else {
-			d(g_assert_not_reached());
-			return;
+		if (!roster_view_find_group (view, &group_iter, name)) {
+			roster_view_group_added (priv->roster, group, view);
+		
+			if (!roster_view_find_group (view, &group_iter, name)) {
+				d(g_assert_not_reached ());
+				return;
+			}
 		}
+
+#if EXPAND
+		if (!gtk_tree_model_iter_has_child (priv->model, &group_iter)) {
+			expand = TRUE;
+		}
+#endif
+
+		gtk_tree_store_append (GTK_TREE_STORE (priv->model),
+				       &iter,
+				       &group_iter);
 	}
 
-#if 0
-	if (!gtk_tree_model_iter_has_child (priv->model, &group_iter)) {
-		/* FIXME: ... */
-		expand = FALSE;
-	}
-#endif
 	e = g_new0 (RosterElement, 1);
 	e->item = gossip_roster_item_ref (item);
-
+ 
 	gtk_tree_store_set (GTK_TREE_STORE (priv->model),
 			    &iter,
 			    COL_IS_GROUP, FALSE,
 			    COL_ELEMENT, e,
 			    -1);
-#if 0
+#if EXPAND
 	if (expand) {
 		GtkTreePath *path;
 
@@ -1542,6 +1979,52 @@ roster_view_add_item (GossipRosterView  *view,
 #endif
 }
 
+static gboolean
+roster_view_remove_item_with_iter (GossipRosterView  *view,
+				   GtkTreeIter       *iter,
+				   GossipRosterItem  *item,
+				   GossipRosterGroup *group)
+{
+	GossipRosterViewPriv *priv;
+	GtkTreeIter           parent;
+	gboolean              ret;
+	gboolean              is_group;
+	RosterElement        *e;
+	gint                  n_children;
+
+	priv = view->priv;
+
+	gtk_tree_model_get (priv->model, iter,
+			    COL_IS_GROUP, &is_group,
+			    COL_ELEMENT, &e,
+			    -1);
+
+	if (is_group) {
+		return TRUE;
+	}
+	
+	if (gtk_tree_model_iter_parent (priv->model, &parent, iter)) {
+		n_children = gtk_tree_model_iter_n_children (priv->model,
+							     &parent);
+	} else {
+		n_children = 0;
+	}
+			
+	g_hash_table_remove (priv->flash_table, item);
+	
+	ret = gtk_tree_store_remove (GTK_TREE_STORE (priv->model), iter);
+
+	gossip_roster_item_unref (e->item);
+	g_free (e);
+
+	/* If the group had one item, it's empty now so remove it. */
+	if (n_children == 1) {
+		roster_view_group_removed (priv->roster, group, view);
+	}
+
+	return ret;
+}
+
 static void
 roster_view_remove_item (GossipRosterView  *view,
 			 GossipRosterItem  *item,
@@ -1549,8 +2032,6 @@ roster_view_remove_item (GossipRosterView  *view,
 {
 	GossipRosterViewPriv *priv;
 	GtkTreeIter           iter;
-	gboolean              is_group;
-	RosterElement        *e;
 
 	priv = view->priv;
 	
@@ -1562,21 +2043,7 @@ roster_view_remove_item (GossipRosterView  *view,
 		return;
 	}
 
-	gtk_tree_model_get (priv->model, &iter,
-			    COL_IS_GROUP, &is_group,
-			    COL_ELEMENT, &e,
-			    -1);
-
-	if (is_group) {
-		return;
-	}
-
-	g_hash_table_remove (priv->flash_table, item);
-
-	gtk_tree_store_remove (GTK_TREE_STORE (priv->model), &iter);
-
-	gossip_roster_item_unref (e->item);
-	g_free (e);
+	roster_view_remove_item_with_iter (view, &iter, item, group);
 }
 
 GossipRosterView *
@@ -1610,14 +2077,6 @@ gossip_roster_view_new (GossipRoster *roster)
 			  "item_removed",
 			  G_CALLBACK (roster_view_item_removed),
 			  view);
-	g_signal_connect (priv->roster,
-			  "group_added",
-			  G_CALLBACK (roster_view_group_added),
-			  view);
-	g_signal_connect (priv->roster,
-			  "group_removed",
-			  G_CALLBACK (roster_view_group_removed),
-			  view);
 
 	return view;
 }
@@ -1691,7 +2150,7 @@ roster_view_flash_timeout_func (FlashTimeoutData *fdata)
 	}
 
 	for (l = gossip_roster_item_get_groups (fdata->item); l; l = l->next) {
-		GossipRosterGroup *group = (GossipRosterGroup *) l->data;
+		GossipRosterGroup *group = l->data;
 
 		if (!roster_view_find_item (view, &iter, fdata->item, group)) {
 			continue;
@@ -1706,18 +2165,26 @@ roster_view_flash_timeout_func (FlashTimeoutData *fdata)
 }
 
 void
-gossip_roster_view_flash_item (GossipRosterView *view,
-			       GossipRosterItem *item,
-			       gboolean          flash)
+gossip_roster_view_flash_contact (GossipRosterView *view,
+				  GossipContact    *contact,
+				  gboolean          flash)
 {
 	GossipRosterViewPriv *priv;
 	FlashData            *flash_data;
+	GossipRosterItem     *item;
 	
 	g_return_if_fail (GOSSIP_IS_ROSTER_VIEW (view));
-	g_return_if_fail (item != NULL);
+	g_return_if_fail (contact != NULL);
 
 	priv = view->priv;
 	
+	item = gossip_roster_get_item (priv->roster, 
+				       gossip_contact_get_jid (contact));
+
+	if (!item) {
+		item = gossip_roster_item_new (gossip_contact_get_jid (contact));
+	}
+	
 	flash_data = g_hash_table_lookup (priv->flash_table, item);
 
 	if (flash && !flash_data) {
@@ -1730,7 +2197,7 @@ gossip_roster_view_flash_item (GossipRosterView *view,
 		flash_data = g_new0 (FlashData, 1);
 		flash_data->flash_on = TRUE;
 		flash_data->flash_timeout_id =
-			g_timeout_add_full (G_PRIORITY_DEFAULT, 350,
+			g_timeout_add_full (G_PRIORITY_DEFAULT, FLASH_TIMEOUT,
 					    (GSourceFunc) roster_view_flash_timeout_func,
 					    fdata,
 					    (GDestroyNotify) roster_view_free_flash_timeout_data);
@@ -1755,23 +2222,32 @@ void
 gossip_roster_view_set_show_offline (GossipRosterView *view,
 				     gboolean show_offline)
 {
+	GossipRoster         *roster;
 	GossipRosterViewPriv *priv;
 	GList                *items, *i;
+	gboolean              show_active;
 
 	g_return_if_fail (GOSSIP_IS_ROSTER_VIEW (view));
 
 	priv = view->priv;
-	
 	priv->show_offline = show_offline;
+	
+	roster = priv->roster;
+	show_active = priv->show_active; /* remember */
 
+	/* disable temporarily */
+	priv->show_active = FALSE;
+	
 	items = gossip_roster_get_all_items (gossip_app_get_roster ());
 	for (i = items; i; i = i->next) {
-		GossipRosterItem *item = (GossipRosterItem *) i->data;
+		GossipRosterItem *item = i->data;
 
-		roster_view_item_presence_updated (gossip_app_get_roster(),
-						   item, view);
+		roster_view_item_presence_updated (roster, item, view);
 	}
 	g_list_free (items);
+
+	/* restore to original setting */
+	priv->show_active = show_active;
 }
 
 gboolean
