@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * Copyright (C) 2003 Imendio HB
+ * Copyright (C) 2003-2004 Imendio HB
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -28,9 +28,11 @@
 #include <libgnome/gnome-i18n.h>
 
 #include "gossip-app.h"
+#include "gossip-contact-list.h"
 #include "gossip-utils.h"
 #include "gossip-marshal.h"
 #include "gossip-roster.h"
+#include "gossip-stock.h"
 
 #define d(x)
 
@@ -44,11 +46,14 @@ struct _GossipRosterPriv {
 	LmConnection     *connection;
         LmMessageHandler *presence_handler;
         LmMessageHandler *iq_handler;
+
+	/* Will eventually replace items and groups hash tables */
+	GHashTable       *contacts;
 };
    
 struct _GossipRosterGroup {
         gchar        *name;
-        GList        *items; /* List of RosterItems */
+        GList        *items;          /* List of RosterItems */
 
 	gint          ref_count;
 };
@@ -59,10 +64,13 @@ struct _GossipRosterItem {
 	gboolean      online;
         gchar        *subscription;
         gchar        *ask;
-        GList        *groups;      /* List of groups */
-        GList        *connections; /* List of RosterConnection */
+        GList        *groups;         /* List of groups */
+        GList        *connections;    /* List of RosterConnection */
 
 	gint          ref_count;
+
+	gboolean      active;         /* If true, is drawn differently
+					 on roster to show change */
 };
    
 typedef struct {
@@ -70,17 +78,17 @@ typedef struct {
 	gint        priority;
 	gchar      *status;
 	GossipShow  show;
-	time_t      last_updated; /* Timestamp to tell when last updated */
+	time_t      last_updated;     /* Timestamp to tell when last updated */
 } RosterConnection;
 
-static void     roster_class_init            (GossipRosterClass *klass);
-static void     roster_init                  (GossipRoster      *roster);
+static void     roster_contact_list_init     (GossipContactListClass *iface);
 static void     roster_finalize              (GObject           *object);
 static void     roster_connected_cb          (GossipApp         *app,
 					      GossipRoster      *roster);
 static void     roster_disconnected_cb       (GossipApp         *app,
 					      GossipRoster      *roster); 
 static void     roster_reset_connection      (GossipRoster      *roster);
+
 static void     roster_clear                 (GossipRoster      *roster);
 static LmHandlerResult
 roster_presence_handler                      (LmMessageHandler  *handler,
@@ -129,6 +137,10 @@ static void     roster_group_remove          (GossipRoster      *roster,
 
 static void     roster_group_free            (GossipRosterGroup *group);
 static gboolean roster_group_is_internal     (GossipRosterGroup *group);
+static GossipPresenceType 
+roster_presence_type_from_show (GossipShow show);
+static GossipContact *
+roster_item_get_contact (GossipRosterItem *item);
  
 /* Signals for item_added, item_removed, ... */
 enum {
@@ -138,41 +150,24 @@ enum {
 	ITEM_PRESENCE_UPDATED,
 	GROUP_ADDED,
 	GROUP_REMOVED,
+
+	/* New GossipContact API */
+	CONTACT_ADDED,
+	CONTACT_REMOVED,
+	CONTACT_UPDATED,
+	CONTACT_PRESENCE_UPDATED,
         LAST_SIGNAL
 };
 
 GObjectClass *parent_class;
 static guint signals[LAST_SIGNAL];
 
-GType
-gossip_roster_get_type (void)
-{
-        static GType object_type = 0;
-         
-        if (!object_type) {
-                static const GTypeInfo object_info = {
-                        sizeof (GossipRosterClass),
-                        NULL,           /* base_init */
-                        NULL,           /* base_finalize */
-                        (GClassInitFunc) roster_class_init,
-                        NULL,           /* class_finalize */
-                        NULL,           /* class_data */
-                        sizeof (GossipRoster),
-                        0,              /* n_preallocs */
-                        (GInstanceInitFunc) roster_init,
-                };
- 
-                object_type = g_type_register_static (G_TYPE_OBJECT,
-                                                      "GossipRoster",
-                                                      &object_info,
-                                                      0);
-        }
- 
-        return object_type;
-}
+G_DEFINE_TYPE_WITH_CODE (GossipRoster, gossip_roster, G_TYPE_OBJECT,
+			 G_IMPLEMENT_INTERFACE (GOSSIP_TYPE_CONTACT_LIST,
+						roster_contact_list_init));
 
 static void     
-roster_class_init (GossipRosterClass *klass)
+gossip_roster_class_init (GossipRosterClass *klass)
 {
         GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
@@ -239,7 +234,7 @@ roster_class_init (GossipRosterClass *klass)
 }
 
 static void
-roster_init (GossipRoster *roster)
+gossip_roster_init (GossipRoster *roster)
 {
         GossipRosterPriv *priv;
 
@@ -253,6 +248,11 @@ roster_init (GossipRoster *roster)
 	priv->groups = g_hash_table_new_full (g_str_hash, g_str_equal,
 					      NULL,
 					      (GDestroyNotify) gossip_roster_group_unref);
+
+	priv->contacts = g_hash_table_new_full (gossip_jid_hash,
+						gossip_jid_equal,
+						(GDestroyNotify) gossip_jid_unref,
+						(GDestroyNotify) gossip_contact_unref);
         
 	priv->connection       = NULL;
         priv->presence_handler = NULL;
@@ -267,6 +267,12 @@ roster_init (GossipRoster *roster)
 			  roster);
 }
 
+static void
+roster_contact_list_init (GossipContactListClass *iface)
+{
+	/* FIXME: Implement */
+}
+
 static void
 roster_finalize (GObject *object)
 {
@@ -376,6 +382,15 @@ roster_clear_foreach_item (gpointer          key,
 	return TRUE;
 }
 
+static gboolean
+roster_clear_foreach_contact (gpointer       key,
+			      GossipContact *contact,
+			      GossipRoster  *roster)
+{
+	g_signal_emit_by_name (roster, "contact_removed", contact);
+	return TRUE;
+}
+
 static gboolean 
 roster_clear_foreach_group (gpointer           key,
 			    GossipRosterGroup *group,
@@ -395,15 +410,21 @@ roster_clear (GossipRoster *roster)
 	/* Go through items and signal item_removed ... */
 	if (priv->items) {
 		g_hash_table_foreach_remove (priv->items,
-				      (GHRFunc) roster_clear_foreach_item,
-				      roster);
+					     (GHRFunc) roster_clear_foreach_item,
+					     roster);
 	}
 
+	if (priv->contacts) {
+		g_hash_table_foreach_remove (priv->contacts,
+					     (GHRFunc) roster_clear_foreach_contact,
+					     roster);
+	}
+	
 	/* Signal group removed */
 	if (priv->groups) {
 		g_hash_table_foreach_remove (priv->groups,
-				      (GHRFunc) roster_clear_foreach_group,
-				      roster);
+					     (GHRFunc) roster_clear_foreach_group,
+					     roster);
 	}
 }
 
@@ -495,11 +516,19 @@ roster_iq_handler (LmMessageHandler *handler,
 		}
 		
 		if (!item) {
+			GossipContact *contact;
+			
 			item = roster_item_new (jid);
-	
+			contact = roster_item_get_contact (item);
+			
 			g_hash_table_insert (priv->items,
 					     gossip_jid_ref (item->jid),
 					     item);
+
+			g_hash_table_insert (priv->contacts,
+					     gossip_jid_ref (item->jid),
+					     contact);
+
 			new_item = TRUE;
 		}
 		
@@ -538,14 +567,18 @@ roster_item_update (GossipRoster     *roster,
 		    LmMessageNode    *node,
 		    gboolean          new_item)
 {
-	const gchar   *subscription;
-	const gchar   *ask;
-	const gchar   *name;
-	LmMessageNode *child;
-	gboolean       in_a_group;
-	GList         *l;
-	GList         *groups;
+	GossipRosterPriv *priv;
+	const gchar      *subscription;
+	const gchar      *ask;
+	const gchar      *name;
+	LmMessageNode    *child;
+	gboolean          in_a_group;
+	GList            *l;
+	GList            *groups;
+	GossipContact    *contact;
 
+	priv = roster->priv;
+	
 	/* Update the item, can be name change, group change, etc.. */
 	subscription = lm_message_node_get_attribute (node, "subscription");
 	if (subscription) {
@@ -558,11 +591,15 @@ roster_item_update (GossipRoster     *roster,
 		g_free (item->ask);
 		item->ask = g_strdup (ask);
 	}
+	
+	contact = g_hash_table_lookup (priv->contacts, item->jid);
 		
 	name = lm_message_node_get_attribute (node, "name");
 	if (name) {
 		g_free (item->name);
 		item->name = g_strdup (name);
+
+		gossip_contact_set_name (contact, item->name);
 	}
 
 	/* FIXME: We should probably check if groups changed before doing
@@ -590,26 +627,53 @@ roster_item_update (GossipRoster     *roster,
 		roster_item_add_to_group (roster, item, UNSORTED_GROUP);
 	}
 
+	
 	if (new_item) {
 		g_signal_emit (roster, signals[ITEM_ADDED], 0, item);
+		g_signal_emit_by_name (roster, "contact_added", contact);
 	} else {
 		g_signal_emit (roster, signals[ITEM_UPDATED], 0, item);
+		g_signal_emit_by_name (roster, "contact_updated", contact);
 	}
 }
 
+static void 
+roster_update_presence (GossipPresence *presence, GossipRosterItem *item)
+{
+	if (gossip_roster_item_is_offline (item)) {
+		gossip_presence_set_state (presence, 
+					   GOSSIP_PRESENCE_STATE_OFFLINE);
+	} else {
+		gossip_presence_set_state (presence,
+					   GOSSIP_PRESENCE_STATE_ONLINE);
+	}
+
+	gossip_presence_set_type (presence,
+				  roster_presence_type_from_show (gossip_roster_item_get_show (item)));
+
+	gossip_presence_set_status (presence, 
+				    gossip_roster_item_get_status (item));
+}
+
 static gboolean
 roster_item_update_presence (GossipRoster     *roster,
 			     GossipRosterItem *item,
 			     GossipJID        *from,
 			     LmMessage        *presence)
 {
+	GossipRosterPriv *priv;
 	LmMessageSubType  type;
 	LmMessageNode    *node;
 	RosterConnection *con = NULL;
 	GList            *l;
+	GossipContact    *contact;
+	GossipPresence   *c_presence;
 	
+	priv = roster->priv;
 	type = lm_message_get_sub_type (presence);
-
+ 
+	contact = g_hash_table_lookup (priv->contacts, item->jid);
+	
 	for (l = item->connections; l; l = l->next) {
 		RosterConnection *connection = (RosterConnection *)l->data;
 		if (gossip_jid_equals (from, connection->jid)) {
@@ -629,7 +693,7 @@ roster_item_update_presence (GossipRoster     *roster,
 	switch (type) {
 	case LM_MESSAGE_SUB_TYPE_AVAILABLE:
 		item->online = TRUE;
-		break;
+	break;
 	case LM_MESSAGE_SUB_TYPE_UNAVAILABLE: 
 		if (con) {
 			roster_item_remove_connection (item, con);
@@ -677,6 +741,11 @@ roster_item_update_presence (GossipRoster     *roster,
 item_updated:
 	g_signal_emit (roster, signals[ITEM_PRESENCE_UPDATED], 0, item);
 
+	c_presence = gossip_contact_get_presence (contact);
+	roster_update_presence (c_presence, item);
+
+	g_signal_emit_by_name (roster, "contact_presence_updated", contact);
+
 	return TRUE;
 }
 
@@ -686,11 +755,14 @@ roster_item_remove (GossipRoster *roster, GossipRosterItem *item)
 	GossipRosterPriv *priv;
 	GList            *l;
 	GList            *groups;
+	GossipContact    *contact;
 	
 	priv = roster->priv;
 
 	gossip_roster_item_ref (item);
+	
 	g_hash_table_remove (priv->items, item->jid);
+
 	g_signal_emit (roster, signals[ITEM_REMOVED], 0, item);
 
 	groups = g_list_copy (item->groups);
@@ -706,6 +778,13 @@ roster_item_remove (GossipRoster *roster, GossipRosterItem *item)
 	g_list_free (item->groups);
 	item->groups = NULL;
 	
+	/* Signal with new GossipContact API */
+	contact = g_hash_table_lookup (priv->contacts, item->jid);
+	gossip_contact_ref (contact);
+	g_hash_table_remove (priv->contacts, item->jid);
+	g_signal_emit_by_name (roster, "contact_removed", contact);
+	gossip_contact_unref (contact);
+
 	gossip_roster_item_unref (item);
 }
 
@@ -874,8 +953,9 @@ roster_group_remove (GossipRoster *roster, GossipRosterGroup *group)
 
 	priv = roster->priv;
 
-	/* FIXME: Not sure if we should remove right away or not... */
-	return;
+	/* FIXME: Not sure if we should remove right away or not... /micke */
+	/* FIXME: I think we should /richard */
+	/*return; */ 
 
 	/* Don't need to worry about the items, if we get here there are no
 	 * items in this groups items list. */
@@ -904,7 +984,51 @@ roster_group_is_internal (GossipRosterGroup *group)
 
 	return FALSE;
 }
- 
+
+static GossipPresenceType
+roster_presence_type_from_show (GossipShow show)
+{
+	switch (show) {
+	case GOSSIP_SHOW_AVAILABLE:
+		return GOSSIP_PRESENCE_TYPE_AVAILABLE;
+	case GOSSIP_SHOW_BUSY:
+		return GOSSIP_PRESENCE_TYPE_BUSY;
+	case GOSSIP_SHOW_AWAY:
+		return GOSSIP_PRESENCE_TYPE_AWAY;
+	case GOSSIP_SHOW_EXT_AWAY:
+		return GOSSIP_PRESENCE_TYPE_EXT_AWAY;
+	default:
+		return GOSSIP_SHOW_AVAILABLE;
+	}
+}
+
+static GossipContact *
+roster_item_get_contact (GossipRosterItem *item)
+{
+	GossipContact       *contact;
+	GossipPresence      *presence; 
+	GossipPresenceState  state;
+	
+	contact = gossip_contact_new_full (GOSSIP_CONTACT_TYPE_CONTACTLIST,
+					   item->jid,
+					   gossip_roster_item_get_name (item));
+	
+	gossip_contact_set_groups (contact, item->groups);
+
+	if (gossip_roster_item_is_offline (item)) {
+		state = GOSSIP_PRESENCE_STATE_OFFLINE;
+	} else {
+		state = GOSSIP_PRESENCE_STATE_ONLINE;
+	}
+
+	presence = gossip_presence_new_full (state,
+					     roster_presence_type_from_show (gossip_roster_item_get_show (item)),
+					     gossip_roster_item_get_status (item));
+	
+	gossip_contact_set_presence (contact, presence);
+
+	return contact;
+} 
 GossipRoster *
 gossip_roster_new (void)
 {
@@ -1024,6 +1148,54 @@ gossip_roster_get_item (GossipRoster *roster,
 	return g_hash_table_lookup (priv->items, jid);
 }
 
+GossipRosterItem *
+gossip_roster_find_item (GossipRoster *roster, const gchar *str)
+{
+	GossipRosterPriv *priv;
+	GossipRosterItem *item = NULL;
+	
+	g_return_val_if_fail (GOSSIP_IS_ROSTER (roster), NULL);
+	g_return_val_if_fail (str != NULL, NULL);
+
+	priv = roster->priv;
+
+	priv = roster->priv;
+	
+	if (gossip_jid_string_is_valid_jid (str)) {
+		GossipJID *jid = gossip_jid_new (str);
+		
+		item = g_hash_table_lookup (priv->items, jid);
+		if (!item) {
+			item = gossip_roster_item_new (jid);
+		}
+	} else {
+		GList *items, *l;
+		gchar *name;
+		
+		name = g_utf8_casefold (str, -1);
+		
+		items = gossip_roster_get_all_items (roster);
+		for (l = items; l; l = l->next) {
+			GossipRosterItem *i = (GossipRosterItem *) l->data;
+			gchar *i_name;
+
+			i_name = g_utf8_casefold (gossip_roster_item_get_name (i), -1);
+
+			if (g_utf8_collate (name, i_name) == 0) {
+				g_free (i_name);
+				item = i;
+				break;
+			}
+			
+			g_free (i_name);
+		}
+		g_free (name);
+		g_list_free (items);
+	}
+
+	return item;
+}
+
 void
 gossip_roster_remove_item (GossipRoster *roster, GossipRosterItem *item)
 {
@@ -1075,6 +1247,7 @@ gossip_roster_rename_item (GossipRoster     *roster,
 	LmMessage        *m;
 	LmMessageNode    *node;
 	const gchar      *jid_str;
+	gchar            *escaped;
 	GList            *l; 
 
 	g_return_if_fail (GOSSIP_IS_ROSTER (roster));
@@ -1095,16 +1268,22 @@ gossip_roster_rename_item (GossipRoster     *roster,
 					"xmlns", "jabber:iq:roster",
 					NULL);
 
+	escaped = g_markup_escape_text (name, -1);
+	
 	node = lm_message_node_add_child (node, "item", NULL);
 	lm_message_node_set_attributes (node, 
 					"jid", jid_str,
-					"name", name,
+					"name", escaped,
 					NULL);
+	g_free (escaped);
+	
 	for (l = item->groups; l; l = l->next) {
-		GossipRosterGroup *group = (GossipRosterGroup *) l->data;
+		GossipRosterGroup *group = l->data;
 
 		if (!roster_group_is_internal (group)) {
-			lm_message_node_add_child (node, "group", group->name);
+			escaped = g_markup_escape_text (group->name, -1);
+			lm_message_node_add_child (node, "group", escaped);
+			g_free (escaped);
 		}
 	}	
 	
@@ -1141,22 +1320,28 @@ gossip_roster_rename_group (GossipRoster      *roster,
 	
 	/* Iterate over all children in the group */
 	for (i = group->items; i; i = i->next) {
-		GossipRosterItem *item = (GossipRosterItem *) i->data;
+		GossipRosterItem *item = i->data;
 		LmMessageNode    *node;
 		const gchar      *jid_str;
+		gchar            *escaped;
 		GList            *l;
 		
 		jid_str = gossip_jid_get_without_resource (item->jid);
+
+		escaped = g_markup_escape_text (item->name, -1);
 		
 		node = lm_message_node_add_child (q_node, "item", NULL);
 		lm_message_node_set_attributes (node,
 						"jid", jid_str,
-						"name", item->name,
+						"name", escaped,
 						NULL);
+
+		g_free (escaped);
 		
 		/* Iterate over all groups in each child */
 		for (l = item->groups; l; l = l->next) {
-			GossipRosterGroup *g = (GossipRosterGroup *) l->data;
+			GossipRosterGroup *g = l->data;
+
 			if (g != group && !roster_group_is_internal (g)) {
 				lm_message_node_add_child (node, 
 							   "group", g->name);
@@ -1170,6 +1355,47 @@ gossip_roster_rename_group (GossipRoster      *roster,
 	lm_message_unref (m);
 }
 
+GossipContact *
+gossip_roster_get_contact_from_item (GossipRoster     *roster,
+				     GossipRosterItem *item)
+{
+	GossipRosterPriv *priv;
+	GossipContact    *contact;
+	
+	g_return_val_if_fail (GOSSIP_IS_ROSTER (roster), NULL);
+
+	priv = roster->priv;
+	
+	contact = g_hash_table_lookup (priv->contacts, item->jid);
+
+	return contact;
+}
+
+const gchar *
+gossip_roster_get_active_resource (GossipRoster  *roster, 
+				   GossipContact *contact) 
+{
+	GossipRosterPriv *priv;
+	GossipRosterItem *item;
+	GossipJID        *jid;
+	GossipJID        *item_jid;
+	
+	g_return_val_if_fail (GOSSIP_IS_ROSTER (roster), "");
+	g_return_val_if_fail (contact != NULL, "");
+	
+	priv = roster->priv; 
+
+	jid = gossip_contact_get_jid (contact);
+	item = g_hash_table_lookup (priv->items, jid);
+
+	if (!item) {
+		return gossip_jid_get_resource (jid);
+	}
+
+	item_jid = gossip_roster_item_get_jid (item);
+
+	return gossip_jid_get_resource (item_jid);
+}
 const gchar * 
 gossip_roster_group_get_name (GossipRosterGroup *group)
 {
@@ -1222,9 +1448,17 @@ gossip_roster_item_get_jid (GossipRosterItem *item)
 const gchar *
 gossip_roster_item_get_name (GossipRosterItem *item)
 {
+	const gchar *name;
+	
 	g_return_val_if_fail (item != NULL, NULL);
 	
-	return item->name;
+	if (!item->name || strcmp (item->name, "") == 0) {
+		name = gossip_jid_get_without_resource (item->jid);
+	} else {
+		name = item->name;
+	}
+	
+	return name;
 }
 
 GossipShow
@@ -1283,6 +1517,23 @@ gossip_roster_item_is_offline (GossipRosterItem *item)
 	return !item->online;
 }
 
+gboolean 
+gossip_roster_item_get_active (GossipRosterItem *item)
+{
+	g_return_val_if_fail (item != NULL, FALSE);
+	
+	return item->active;
+}
+
+void
+gossip_roster_item_set_active (GossipRosterItem *item,
+			       gboolean          active)
+{
+	g_return_if_fail (item != NULL);
+
+	item->active = active;
+}
+
 GossipRosterItem *
 gossip_roster_item_ref (GossipRosterItem *item)
 {
